import numpy as np
import scipy.linalg as lin
import scipy.optimize as opt


def optimal_fusion(mean_a, mean_b, cov_a, cov_b, cov_ab):
    """Fuse two estimates using optimal fusion rule.

    Parameters
    ----------
    mean_a : ndarray (n,)
        First unbiased input estimate.
    mean_b : ndarray (n,)
        Second unbiased input estimate.
    cov_a : ndarray (n,n)
        Bound on covariance matrix of first input estimate.
    cov_b : ndarray (n,n)
        Bound on covariance matrix of second input estimate.
    cov_ab : ndarray(n,n)
        Cross-covariance between estimates a and b.

    Returns
    -------
    mean_c : ndarray (n,)
        Fused (unbiased) estimate.
    cov_c : ndarray (n,n)
        Bound on covariance matrix of fused estimate.

    """
    temp = np.linalg.inv(cov_a + cov_b - cov_ab - cov_ab.T)
    mean_c = mean_a + (cov_a - cov_ab) @ temp @ (mean_b - mean_a)
    cov_c = cov_a - (cov_a - cov_ab) @ temp @ (cov_a - cov_ab.T)

    return mean_c, cov_c


def covariance_intersection(mean_a, mean_b, cov_a, cov_b):
    """Fuse two estimates using covariance intersection.

    Parameters
    ----------
    mean_a : ndarray (n,)
        First unbiased input estimate.
    mean_b : ndarray (n,)
        Second unbiased input estimate.
    cov_a : ndarray (n,n)
        Bound on covariance matrix of first input estimate.
    cov_b : ndarray (n,n)
        Bound on covariance matrix of second input estimate.

    Returns
    -------
    mean_c : ndarray (n,)
        Fused (unbiased) estimate.
    cov_c : ndarray (n,n)
        Bound on covariance matrix of fused estimate.
    w : float
        Weight used for fusion (first input estimate).

    """

    # Minimize trace of fused covariance matrix to find weight w
    inf_a = np.linalg.inv(cov_a)
    inf_b = np.linalg.inv(cov_b)
    cost = lambda w: np.trace(np.linalg.inv(w * inf_a + (1 - w) * inf_b))
    jac = lambda w: -np.trace(
        np.linalg.inv(w * inf_a + (1 - w) * inf_b)
        @ (inf_a - inf_b)
        @ np.linalg.inv(w * inf_a + (1 - w) * inf_b)
    )
    result = opt.minimize(cost, 0.5, jac=jac, bounds=[(0.0, 1.0)])
    if result.success:
        w = result.x
    else:
        raise ValueError("optimizer did not converge")

    # Compute conservative covariance and mean
    cov_c = np.linalg.inv(w * inf_a + (1 - w) * inf_b)
    mean_c = cov_c @ (w * inf_a @ mean_a + (1 - w) * inf_b @ mean_b)

    return mean_c, cov_c, w


def random_mean_quant(mean, cov, delta, x_max):
    """Quantize mean such that result is unbiased.

    Parameters
    ----------
    mean : ndarray (n,)
        Input vector to be quantized.
    cov : ndarray (n,n)
        Error covariance matrix of input vector.
    delta : float
        Resolution for quantization.
    x_max : float
        Maximum codeword.

    Returns
    -------
    mean_quant : ndarray (n,)
        Quantized input vector.
    cov_bound : ndarray (n,n)
        Upper bound on covariance matrix of quantized input vector.

    """
    # Quantize mean with uniform dither
    mean_quant = (
        np.rint(mean / delta + np.random.uniform(-0.5, 0.5, len(mean))) * delta + 0.0
    )

    # Raise error if quantizer saturates
    if np.any(mean_quant > x_max):
        raise ValueError("estimate quantizer saturated")
    if np.any(mean_quant < -x_max + delta):
        raise ValueError("estimate quantizer saturated")

    # Make covariance matrix consistent
    cov_bound = cov + delta ** 2 * np.eye(len(cov))

    return mean_quant, cov_bound


def ddom_cov_quant(cov, delta_diag, delta_offd, x_max):
    """Conservatively quantize covariance matrix using diagonal dominance criterion.

    Parameters
    ----------
    cov : ndarray (n,n)
        Covariance matrix to be quantized.
    delta_diag : float
        Resolution for quantization of diagonal elements.
    delta_diag : float
        Resolution for quantization of off-diagonal elements.
    x_max : float
        Maximum off-diagonal codeword.

    Returns
    -------
    cov_quant : ndarray (n,n)
        Conservatively quantized input covariance matrix.

    """
    # Quantize off diagonal elements by rounding to the nearest codeword
    cov_quant = np.rint(cov / delta_offd) * delta_offd
    cov_offd_error = cov_quant - cov
    np.fill_diagonal(cov_offd_error, 0.0)

    # Compute off-diagonal quantization error absolute sums
    offd_error_sums = np.sum(np.abs(cov_offd_error), axis=1)

    # Make quantization error diagonally dominant (Gershgorin circles in RHP)
    cov_quant[np.diag_indices_from(cov_quant)] = (
        np.ceil((np.diag(cov) + offd_error_sums) / delta_diag) * delta_diag
    )

    # Raise error if quantizer saturates
    cov_quant_offd = cov_quant - np.diag(np.diag(cov_quant))
    if np.any(cov_quant_offd > x_max):
        raise ValueError("off-diagonal quantizer saturated")
    if np.any(cov_quant_offd < (-x_max + delta_offd)):
        raise ValueError("off-diagonal quantizer saturated")
    cov_quant_diag = np.diag(cov_quant)
    if np.any(cov_quant_diag > (x_max + (cov.shape[0] - 1) * delta_offd / 2)):
        raise ValueError("diagonal quantizer saturated")

    return cov_quant


def mchol_cov_quant(cov, delta_diag, delta_offd, x_max):
    """Quantize covariance matrix using modified Cholesky decomposition such that result is conservative.

    Parameters
    ----------
    cov : ndarray (n,n)
        Covariance matrix to be quantized.
    delta_diag : float
        Resolution for quantization of diagonal elements.
    delta_diag : float
        Resolution for quantization of off-diagonal elements.
    x_max : float
        Maximum off-diagonal codeword.

    Returns
    -------
    cov_quant : ndarray (n,n)
        Conservatively quantized input covariance matrix.

    """
    # Round off-diagonal part of matrix and compute quantization error
    cov_diag = np.diag(cov)
    cov_offd = cov - np.diag(cov_diag)
    cov_offd_round = np.rint(cov_offd / delta_offd) * delta_offd + 0.0
    cov_err_old = np.diag(cov_diag) + cov_offd_round - cov
    cov_err = np.copy(cov_err_old)
    n = len(cov_err)

    # Compute lower bounds for Gershgorin lower bounds for each row/column
    G = np.zeros(n)
    for i in range(n):
        G[i] = (
            cov_err[i, i]
            - np.linalg.norm(cov_err[:i, i], 1)
            - np.linalg.norm(cov_err[i + 1 :, i], 1)
        )

    E = np.zeros(n)
    P = list(range(n))
    for j in range(n):
        # Perform pivoting by maximum lower bound of Gershgorin bound (-> lower perturbation E).
        idx = np.argmax(G[j:]) + j
        G[j], G[idx] = G[idx], G[j]
        P[j], P[idx] = P[idx], P[j]
        foo = np.copy(cov_err[j, :])
        cov_err[j, :] = cov_err[idx, :]
        cov_err[idx, :] = foo
        foo = np.copy(cov_err[:, j])
        cov_err[:, j] = cov_err[:, idx]
        cov_err[:, idx] = foo

        if np.any(cov_err[j:, j] != 0.0):
            # Compute shift E_jj.
            a_norm = np.linalg.norm(cov_err[j + 1 :, j], 1)
            E[j] = max(0.0, a_norm - cov_err[j, j])
            e_old = E[j]
            cov_err[j, j] += E[j]

            # If this is the last diagonal element of the matrix the rest of the loop can be skipped.
            if j == len(cov) - 1:
                break

            # Update lower bounds of Gershgorin lower bounds of the remaining rows/columns.
            temp = 1 - a_norm / cov_err[j, j]
            G[j + 1 :] += cov_err[j + 1 :, j] * temp

            # Perform Cholesky decomposition step for current submatrix as usual.
            cov_err[j, j] = np.sqrt(cov_err[j, j])
            cov_err[j + 1 :, j] /= cov_err[j, j]
            cov_err[j + 1 :, j + 1 :] -= np.outer(
                cov_err[j + 1 :, j], cov_err[j + 1 :, j]
            )

    # Compute inverse permutation and round diagonal elements
    P_inv = np.argsort(P)
    cov_diag_round = np.ceil((cov_diag + E[P_inv]) / delta_diag) * delta_diag
    cov_quant = np.diag(cov_diag_round) + cov_offd_round

    # Saturate with warning if off-diagonal scalar is too large
    idx = (cov_quant - np.diag(np.diag(cov_quant))) > x_max
    if np.any(idx):
        raise ValueError("off-diagonal quantizer saturated")
    idx = (cov_quant - np.diag(np.diag(cov_quant))) < (-x_max + delta_offd)
    if np.any(idx):
        raise ValueError("off-diagonal quantizer saturated")
    idx = np.diag(cov_quant) > (x_max + (cov.shape[0] - 1) * delta_offd / 2)
    if np.any(idx):
        raise ValueError("diagonal quantizer saturated")

    return cov_quant